/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.id.XID;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContainer;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeID;
import org.deckfour.xes.model.XAttributeList;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.mapdb.BTreeMap;
import org.processmining.xeslite.external.MapDBStore.Builder;

/**
 * Factory storing attributes of {@link XAttributable} XES objects in a,
 * possibly external, {@link ExternalStore}. This implementation stores
 * attributes of {@link XEvent}, {@link XTrace} objects in a {@link DiskStore}
 * on the file system using MapDB as storage provider. Attributes are stored
 * together in a single {@link BTreeMap}.
 * <p>
 * Use one of the embedded sub-classes in this class to create a version that
 * fits your needs. Currently, there are:
 * <ul>
 * <li> {@link MapDBDiskImpl}
 * <li> {@link MapDBDiskWithoutCacheImpl}
 * <li> {@link MapDBDiskSequentialAccessImpl}
 * <li> {@link MapDBDiskSequentialAccessWithoutCacheImpl}
 * </ul>
 * <p>
 * PLEASE NOTE: Attributes are immutable in this implementation, as they may be
 * serialized and re-created at any time. Trying to change the value of an
 * attribute for example through the {@link XAttributeLiteral#setValue(String)}
 * method will result in an {@link UnsupportedOperationException}. To change the
 * value of an existing attribute, please just use the
 * {@link XAttributeMap#put(String, org.deckfour.xes.model.XAttribute)} method.
 * 
 * @author F. Mannhardt
 * 
 */
public abstract class XFactoryExternalStore implements XFactory {

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk.
	 * <p>
	 * This version does store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory. This implementation
	 * is NOT guaranteed to be synchronized!
	 * 
	 * @author F. Mannhardt
	 * 
	 */
	public static class MapDBDiskImpl extends XFactoryExternalStore {

		public static void register() {
			if (!containsFactory(MapDBDiskImpl.class)) {
				XFactoryRegistry.instance().register(new MapDBDiskImpl());	
			}
		}

		private final ExternalStore attributeStore;

		public MapDBDiskImpl() {
			this(new MapDBStore.Builder());
		}

		public MapDBDiskImpl(MapDBDatabase database) {
			this(new MapDBStore.Builder().withDatabase(database));
		}

		public MapDBDiskImpl(MapDBStore.Builder dbBuilder) {
			super();
			attributeStore = dbBuilder.build();
		}

		@Override
		protected final ExternalStore getStore() {
			return attributeStore;
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for speed";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version does store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

	}

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk. This version
	 * does <b>NOT</b> store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory.
	 * 
	 * @author F. Mannhardt
	 * 
	 */
	public static class MapDBDiskWithoutCacheImpl extends MapDBDiskImpl {

		public static void register() {
			if (!containsFactory(MapDBDiskWithoutCacheImpl.class)) {
				XFactoryRegistry.instance().register(new MapDBDiskWithoutCacheImpl());	
			}
		}

		public MapDBDiskWithoutCacheImpl() {
			super();
		}

		public MapDBDiskWithoutCacheImpl(Builder dbBuilder) {
			super(dbBuilder);
		}

		public MapDBDiskWithoutCacheImpl(MapDBDatabase database) {
			super(database);
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for memory";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version does NOT store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

		@Override
		public XEvent createEvent() {
			return new XEventBareExternalImpl(getStore());
		}

		@Override
		public XEvent createEvent(XAttributeMap attributes) {
			return new XEventBareExternalImpl(attributes, getStore());
		}

	}

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk. This version is
	 * optimized for sequential access of events, performance will degraded when
	 * using random-access methods of XTrace.
	 * <p>
	 * This version does store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory.
	 * 
	 * @author F. Mannhardt
	 * 
	 */
	public static class MapDBDiskSequentialAccessImpl extends MapDBDiskImpl {

		public static void register() {
			if (!containsFactory(MapDBDiskSequentialAccessImpl.class)) {
				XFactoryRegistry.instance().register(new MapDBDiskSequentialAccessImpl());	
			}
		}
		
		public MapDBDiskSequentialAccessImpl() {
			super();
		}

		public MapDBDiskSequentialAccessImpl(Builder dbBuilder) {
			super(dbBuilder);
		}

		public MapDBDiskSequentialAccessImpl(MapDBDatabase database) {
			super(database);
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for speed & sequential access";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version is optimized for sequential access of events, performance will degraded when using random-access methods of XTrace.  This version does store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

		@Override
		public XTrace createTrace() {
			return new XTraceCompressedExternalImpl(true, getStore());
		}

		@Override
		public XTrace createTrace(XAttributeMap attributes) {
			return new XTraceCompressedExternalImpl(true, attributes, getStore());
		}

		@Override
		public XTrace createTrace(Collection<XEvent> events) {
			return new XTraceCompressedExternalImpl(true, getStore(), events);
		}

		@Override
		public XEvent createEvent() {
			return new XEventCachingExternalImpl(getStore());
		}

		@Override
		public XEvent createEvent(XAttributeMap attributes) {
			return new XEventCachingExternalImpl(attributes, getStore());
		}

	}

	/**
	 * A XES Factory that stores XAttributes with MapDB on disk. This version is
	 * optimized for sequential access of events, performance will degraded when
	 * using random-access methods of XTrace.
	 * <p>
	 * This version does <b>NOT</b> store common attributes like 'concept:name',
	 * 'lifecycle:transition', 'time:timestamp' in memory.
	 * 
	 * @author F. Mannhardt
	 * 
	 */
	public static class MapDBDiskSequentialAccessWithoutCacheImpl extends MapDBDiskSequentialAccessImpl {

		public static void register() {
			if (!containsFactory(MapDBDiskSequentialAccessWithoutCacheImpl.class)) {
				XFactoryRegistry.instance().register(new MapDBDiskSequentialAccessWithoutCacheImpl());	
			}
		}
		
		public MapDBDiskSequentialAccessWithoutCacheImpl() {
			super();
		}

		public MapDBDiskSequentialAccessWithoutCacheImpl(Builder dbBuilder) {
			super(dbBuilder);
		}

		public MapDBDiskSequentialAccessWithoutCacheImpl(MapDBDatabase database) {
			super(database);
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for memory & sequential access";
		}

		@Override
		public String getDescription() {
			return "A XES Factory that stores XAttributes with MapDB on disk. This version is optimized for sequential access of events, performance will degraded when using random-access methods of XTrace. This version does NOT store common attributes like 'concept:name', 'lifecycle:transition', 'time:timestamp' in memory. This implementation is NOT guaranteed to be synchronized!";
		}

		@Override
		public XTrace createTrace() {
			return new XTraceCompressedExternalImpl(false, getStore());
		}

		@Override
		public XTrace createTrace(XAttributeMap attributes) {
			return new XTraceCompressedExternalImpl(false, attributes, getStore());
		}

		@Override
		public XTrace createTrace(Collection<XEvent> events) {
			return new XTraceCompressedExternalImpl(false, getStore(), events);
		}

		@Override
		public XEvent createEvent() {
			return new XEventBareExternalImpl(getStore());
		}

		@Override
		public XEvent createEvent(XAttributeMap attributes) {
			return new XEventBareExternalImpl(attributes, getStore());
		}

	}
	
	public static class MapDBDiskOptimizedForWritingImpl extends MapDBDiskWithoutCacheImpl {

		public MapDBDiskOptimizedForWritingImpl() {
			super(new MapDBDatabaseImpl().setAsyncWriteQueueSize(32 * 1024));
		}

		public MapDBDiskOptimizedForWritingImpl(Builder dbBuilder) {
			super(dbBuilder);
		}

		public MapDBDiskOptimizedForWritingImpl(MapDBDatabase database) {
			super(database);
		}

		@Override
		public String getName() {
			return "XESLite: MapDB Optimized for writing concurrently";
		}

	}

	protected abstract ExternalStore getStore();

	public static boolean containsFactory(Class<? extends XFactoryExternalStore> clazz) {
		for (XFactory f: XFactoryRegistry.instance().getAvailable()) {
			if (clazz == f.getClass()) {
				return true;
			}
		}
		return false;
	}

	private final Integer intern(String key) {
		return getStore().getAttributeKeyPool().put(key);
	}

	@Override
	public String toString() {
		return getName();
	}

	@Override
	public String getAuthor() {
		return "F. Mannhardt";
	}

	@Override
	public URI getUri() {
		return URI.create("http://www.xes-standard.org/");
	}

	@Override
	public String getVendor() {
		return "xes-standard.org";
	}

	@Override
	public XLog createLog() {
		return new XLogExternalImpl(getStore());
	}

	@Override
	public XLog createLog(XAttributeMap attributes) {
		return new XLogExternalImpl(attributes, getStore());
	}

	public XLog createLog(Collection<XTrace> traces) {
		return new XLogExternalImpl(getStore(), traces);
	}

	@Override
	public XTrace createTrace() {
		return new XTraceExternalImpl(getStore());
	}

	public XTrace createTrace(Collection<XEvent> events) {
		return new XTraceExternalImpl(getStore(), events);
	}

	@Override
	public XTrace createTrace(XAttributeMap attributes) {
		return new XTraceExternalImpl(attributes, getStore());
	}

	@Override
	public XEvent createEvent() {
		return new XEventCachingExternalImpl(getStore());
	}

	@Override
	public XEvent createEvent(XAttributeMap attributes) {
		return new XEventCachingExternalImpl(attributes, getStore());
	}

	@Override
	public XEvent createEvent(XID id, XAttributeMap attributes) {
		throw new UnsupportedOperationException("Can not create an XEvent with a pre-defined ID");
	}

	@Override
	public XAttributeMap createAttributeMap() {
		return new XAttributeMapImpl();
	}

	@Override
	public XAttributeBoolean createAttributeBoolean(String key, boolean value, XExtension extension) {
		return new XAttributeBooleanExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeContinuous createAttributeContinuous(String key, double value, XExtension extension) {
		return new XAttributeContinuousExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeDiscrete createAttributeDiscrete(String key, long value, XExtension extension) {
		return new XAttributeDiscreteExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeLiteral createAttributeLiteral(String key, String value, XExtension extension) {
		return new XAttributeLiteralExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeTimestamp createAttributeTimestamp(String key, Date value, XExtension extension) {
		return new XAttributeTimestampExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeTimestamp createAttributeTimestamp(String key, long millis, XExtension extension) {
		return new XAttributeTimestampExternalImpl(intern(key), millis, extension, getStore());
	}

	@Override
	public XAttributeID createAttributeID(String key, XID value, XExtension extension) {
		return new XAttributeIDExternalImpl(intern(key), value, extension, getStore());
	}

	@Override
	public XAttributeContainer createAttributeContainer(String key, XExtension extension) {
		return new XAttributeContainerExternalImpl(intern(key), extension, getStore());
	}

	@Override
	public XAttributeList createAttributeList(String key, XExtension extension) {
		return new XAttributeListExternalImpl(intern(key), extension, getStore());
	}

	public void startPump() {
		if (getStore().isPumping()) {
			throw new IllegalStateException("Pump has already been started!");
		}
		getStore().startPump();
	}

	public boolean isPumping() {
		return getStore().isPumping();
	}

	public final XEvent pumpEvent(XEvent pumpEvent) {
		if (pumpEvent.hasAttributes()) {
			List<XAttribute> pumpValues = (List<XAttribute>) pumpEvent.getAttributes().values();
			pumpMetaAttributes(pumpValues);
			XEvent event = createEvent();
			getStore().getPumpService().pumpAttributes(event, pumpValues);
			return event;
		} else {
			return createEvent();
		}
	}

	public final XTrace pumpTrace(XTrace pumpTrace) {
		if (pumpTrace.hasAttributes()) {
			// Assumes events already have been pumped, only handle our own attributes
			List<XAttribute> pumpValues = (List<XAttribute>) pumpTrace.getAttributes().values();
			pumpMetaAttributes(pumpValues);
			XTrace trace = createTrace(pumpTrace);
			getStore().getPumpService().pumpAttributes(trace, pumpValues);
			return trace;
		} else {
			return createTrace(pumpTrace);
		}
	}

	public final XLog pumpLog(XLog pumpLog) {
		if (pumpLog.hasAttributes()) {
			// Assumes traces already have been pumped, only handle our own attributes
			List<XAttribute> pumpValues = (List<XAttribute>) pumpLog.getAttributes().values();
			pumpMetaAttributes(pumpValues);
			XLog log = createLog(pumpLog);
			copyLogMetadata(pumpLog, log);
			getStore().getPumpService().pumpAttributes(log, pumpValues);
			return log;
		} else {
			XLog log = createLog(pumpLog);
			copyLogMetadata(pumpLog, log);
			return log;
		}
	}

	private static void copyLogMetadata(XLog source, XLog target) {
		target.getClassifiers().addAll(source.getClassifiers());
		target.getExtensions().addAll(source.getExtensions());
		target.getGlobalEventAttributes().addAll(source.getGlobalEventAttributes());
		target.getGlobalTraceAttributes().addAll(source.getGlobalTraceAttributes());
	}

	private final void pumpMetaAttributes(List<XAttribute> attributes) {
		int i = 0;
		for (XAttribute a : attributes) {
			if (a.hasAttributes()) {
				attributes.set(i, pumpAttribute(a));
			}
			i++;
		}
	}
	
	private final XAttribute pumpAttribute(XAttribute pumpAttribute) {
		if (pumpAttribute.hasAttributes()) {
			// First handle meta attributes
			Collection<XAttribute> values = pumpAttribute.getAttributes().values();
			List<XAttribute> metaAttributes = new ArrayList<>(values.size());
			for (XAttribute metaAttribute : values) {
				metaAttributes.add(pumpAttribute(metaAttribute));
			}
			ExternalAttribute attribute = XAttributeExternalImpl.convert(getStore(), pumpAttribute);
			attribute.setExternalId(getStore().getIdFactory().nextId());
			getStore().getPumpService().pumpAttributes(attribute, metaAttributes);
			return attribute;
		} else {
			return pumpAttribute;
		}
	}

	public void finishPump() throws InterruptedException {
		getStore().getPumpService().finishPump();
	}
	
	public XFactory createPumpTransferFactory() {
		return new PumpFactoryImpl(getStore());
	}
	
	public void dispose() {
		getStore().dispose();
	}

}
